from collections import UserDict
from datetime import datetime, date
import json
import re


class WrongPhone(Exception):
    pass


class WrongDate(Exception):
    pass


class WrongEmail(Exception):
    pass


class WrongDays(Exception):
    pass


class NoName(Exception):
    pass


class Field:
    def __init__(self, value):
        self._value = value

    def __str__(self):
        return str(self._value)

    def __repr__(self):
        return str(self)

    def __eq__(self, other):
        if self.__class__ == other.__class__:
            return self._value == other.value
        return False

    # Додав магічний метод __hash__ тому, що діти класу Field
    # можуть бути елементами множини(set)
    def __hash__(self):
        return hash(self._value)


class Name(Field):
    @staticmethod
    def is_valid_name(name):
        if name.strip() == '':
            return False
        else:
            return True

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, val):
        if self.is_valid_name(val):
            self._value = val
        else:
            raise NoName("You missed to enter name. "
                             "Please check the value and try again")


class Phone(Field):

    @staticmethod
    def is_valid_phone(phone):
        if phone == '':
            return True
        # Валідація номеру телефону відбувається за
        # допомогою регулярнго виразу, що означає: необов'язково +, потім цифра від 1 до 9
        # та 11 цифр від 0 до 9. Тобто валідними будуть такі номери +123456987456 та 123456987456,
        match = re.search(r"^\+?[1-9][\d]{11}$", phone)
        return bool(match)

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, val):
        if self.is_valid_phone(val):
            self._value = val
        else:
            raise WrongPhone("You tried to enter an invalid phone number. "
                             "Please check the value and try again")


class Birthday(Field):

    @staticmethod
    def is_valid_date(date):
        # Валідація дня народження є значно простішою, ніж номеру телефону.
        # Якщо ввід користувача перетворюється у об'єкт datetime то дата валідна
        try:
            if date == '':
                return True
            datetime.strptime(str(date), "%d.%m.%Y")
            return True
        except ValueError:
            return False

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, val):
        if self.is_valid_date(val):
            self._value = val
        else:
            raise WrongDate(
                "Invalid date. Please enter birthday in format 'DD.MM.YYYY'.")


class Address:
    def __init__(self, street="", city="", country="", postcode=""):
        self.street = street
        self.city = city
        self.country = country
        self.postcode = postcode

    def __str__(self):
        return f"{self.street},{self.city},{self.country},{self.postcode}"

    def __repr__(self):
        return f"Address(street='{self.street}', city='{self.city}', country='{self.country}', postcode='{self.postcode}')"


class Email:
    def __init__(self, value=""):
        self.value = value

    @staticmethod
    def is_valid_email(email):
        if email == '':
            return True
        match = re.search(
            r"^[a-zA-Z0-9._%+-]+@[a-zA-Z0-9.-]+\.[a-zA-Z]{2,}$", email)
        return bool(match)

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, val):
        if self.is_valid_email(val):
            self._value = val

        else:
            raise WrongEmail(
                "Invalid email address. Please enter a correct email address")

    def __str__(self) -> str:
        return self.value

    def __repr__(self) -> str:
        return str(self)


class Record:
    def __init__(self, name, phones=None, birthday=None, address=None, email=None):
        self.name = name
        self.phones = phones
        self.birthday = birthday
        self.address = address
        self.email = email

    def __str__(self):
        # Рядкове представлення Record у форматі
        # Володя: +123456987456, 23456987456. Birthday: 21.01.1978
        phones = ", ".join([str(phone) for phone in self.phones])
        birthday = f"Birthday: {self.birthday}" if self.birthday.value else "Birthday:"
        address_str = f"Address: {self.address}" if self.address else ""
        email_str = f"Email: {self.email}" if self.email else ""
        return f"{self.name.value}: {phones}, {birthday}, {address_str}, {email_str}"

    def __repr__(self):
        return str(self)

    def add_phone(self, phone: Phone):
        if phone in self.phones:
            return f"User {self.name.value} already has {phone} phone number."
        else:
            self.phones.append(phone)
        # Список телефонів приводиться до множини для того, щоб виключити можливість
        # повторення номеру телефону
        self.phones = list(set(self.phones))
        return f"Phone number {phone} is added successfully for user {self.name.value}."

    def change_phone(self, old_number: Phone, new_number: Phone):
        # У списку телефонів знаходиться індекс старого номера та змінює
        # old_number на new_number
        if old_number not in self.phones:
            return f"Number {old_number} not found."
        else:
            phone_number_index = self.phones.index(old_number)
            self.phones[phone_number_index] = new_number
            return f"The phone number {old_number} for the user {self.name} "\
                f"has been changed to {new_number}"

    def delete_phone(self, phone):
        try:
            self.phones.remove(phone)
            return f"Phone number {phone} for user {self.name} deleted successfully."
        except ValueError:
            return f"Phone number {phone} for user {self.name} not found"

    def days_to_birthday(self):
        try:
            birthday = datetime.strptime(str(self.birthday), "%d.%m.%Y").date()
        except ValueError:
            return f"No birthday for user {self.name.value}"

        today = date.today()
        birthday = birthday.replace(year=today.year)
        # Якщо цього року вже був день народження, кількість днів рахується до наступного
        # дня народження
        if birthday < today:
            birthday = birthday.replace(year=today.year+1)
        result = (birthday - today).days
        birthday_str = birthday.strftime("%d %B")
        return f"The birthday of user {self.name} will be in {result} days, {birthday_str}"

    def change_birthday(self, birthday: Birthday):
        self.birthday = birthday
        return f"Birthday date for user {self.name.value} is changed to {birthday} successfully."

    def change_address(self, address: Address):
        self.address = address
        return f"Address {address} for user {self.name.value} is changed successfully."

    def change_email(self, new_email: Email):
        self.email = new_email
        return f"Email {new_email} for user {self.name.value} is changed successfully."


class AddressBook(UserDict):
    def add_record(self, record):
        self[record.name.value] = record

    def delete_record(self, name: Name):
        del self[name.value]

    def change_record(self, name, new_record):
        self[new_record.name.value] = new_record

    def search(self, field: str, text: str) -> list[Record]:
        result = []
        if field.lower() == "name":
            for record in self.data.values():
                if text in record.name.value:
                    result.append(record)
        elif field.lower() == "phone":
            for record in self.data.values():
                for phone in record.phones:
                    if text in phone.value:
                        result.append(record)
        elif field.lower() == "email":
            for record in self.data.values():
                if text.lower() in record.email.value.lower():
                    result.append(record)
        return result

    def show_birthday(self, days: int):
        result_list = []
        result_to_print = []
        now = datetime.today()
        start_date = now.replace(hour=0, minute=0, second=0, microsecond=0)
        for record in self.data.values():
            if record.birthday.value:
                birthday = datetime.strptime(str(record.birthday.value), "%d.%m.%Y")
                birthday = birthday.replace(year=start_date.year)
                if birthday < start_date:
                    birthday = birthday.replace(year=start_date.year+1)
                gap = birthday - start_date
                gap = gap.days
                result_list.append([str(record.name.value), birthday, gap])
        result_list = sorted(result_list, key=lambda person: person[2])
        result_to_print = [f'{value[1].strftime("%d.%m.%Y")}: {value[0]}' for value in result_list if value[2] <= days]
        if result_to_print:
            print(f"Birthdays today and within {days} next day(s):")
            print(*result_to_print, sep='\n')
        else:
            if days == 1:
                return f"There are no birthdays to show within {days} day"
            else:
                return f"There are no birthdays to show within {days} days"
        return None

    @classmethod
    def open_file(cls, filename):
        """Take as input filename. Return AddressBook"""
        try:
            with open(filename, encoding="utf-8") as file:
                json_data = json.load(file)
                data = cls()
                for name, record in json_data.items():
                    name = Name(name)
                    phones = [Phone(phone) for phone in record["phones"]]
                    birthday = Birthday(record["birthday"])
                    address = Address(**record["address"])
                    email = Email(record["email"])

                    data.add_record(
                        Record(name, phones, birthday, address, email))
        except FileNotFoundError:
            data = cls()
        return data

    def write_to_file(self, filename: str):
        json_data = {}

        for name, record in self.data.items():
            json_data[name] = {
                "phones": [phone.value for phone in record.phones],
                "birthday": record.birthday.value if record.birthday is not None else '',
                "address": {
                    "street": record.address.street if record.address is not None else '',
                    "city": record.address.city if record.address is not None else '',
                    "country": record.address.country if record.address is not None else '',
                    "postcode": record.address.postcode if record.address is not None else '',
                },
                "email": record.email.value if record.email is not None else ''
            }
        with open(filename, "w", encoding="utf-8") as file:
            json.dump(json_data, file, indent=4, ensure_ascii=False)

    # Методи __iter__ та __next__ перетворюють елземпляри AddressBook на
    # ітератори, щоб користувачам показувати одночасно self.page_size записів

    def __iter__(self):
        self.current_page = 1
        self.page_size = 10
        self.start_index = (self.current_page - 1) * self.page_size
        self.end_index = self.start_index + self.page_size
        return self

    def __next__(self):
        if self.start_index >= len(self.data):
            raise StopIteration

        page_records = list(self.data.values())[
            self.start_index:self.end_index]
        self.start_index = self.end_index
        self.end_index = self.start_index + self.page_size
        self.current_page += 1

        return page_records


if __name__ == "__main__":
    pass



import sys
import shutil
import os
from pathlib import Path

CATEGORIES = {
    'archives': ['.rar', '.zip'],
    'audio': ['.mp3', '.wav'],
    'documents': ['.docx', '.pptx'],
    'video': ['.avi', '.mp4'],
    'images': ['.jpeg', '.png']
}


def move_file(path: Path, root_dir: Path, cat: str) -> None:
    target_dir = root_dir.joinpath(cat)
    if not target_dir.exists():
        target_dir.mkdir()
    new_name = target_dir.joinpath(f"{(path.stem)}{path.suffix}")
    path.replace(new_name)


def get_categories(path: Path) -> str:
    ext = path.suffix.lower()
    for cat, exts in CATEGORIES.items():
        if ext in exts:
            return cat
    return 'Other'


def sort_folder(path: Path) -> None:
    ignore_path = []

    for i in CATEGORIES:
        path_to_ignore = path.joinpath(i)
        ignore_path.append(path_to_ignore)

    for item in path.glob('**/*'):
        need_to_ignore = False
        if item.is_file():
            cat = get_categories(item)
            for check in ignore_path:
                if check not in list(item.parents):
                    continue
                else:
                    need_to_ignore = True
                    break
            if need_to_ignore is False:
                move_file(item, path, cat)
    for item in path.glob('**/*'):
        if item.is_dir():
            new_path = item.parent.joinpath((item.name))
            item.replace(new_path)


def unpack_archive(path: Path):
    for item in path.glob('**/*'):
        if item.is_file():
            cat = get_categories(item)
            if cat == 'archives':
                new_path = path.joinpath(cat).joinpath(item.stem)
                shutil.unpack_archive(item, new_path)


def removeEmptyFolders(path: Path, removeRoot=True):
    if not os.path.isdir(path):
        return

    # remove empty subfolders
    files = os.listdir(path)
    if len(files):
        for f in files:
            fullpath = os.path.join(path, f)
            if os.path.isdir(fullpath):
                removeEmptyFolders(fullpath)

    # if folder empty, delete it
    files = os.listdir(path)
    if len(files) == 0 and removeRoot:
        print("Removing empty folder:", path)
        os.rmdir(path)


def get_results(path: Path):
    know_suffix = []
    unknown_suffix = []
    for item in path.glob('*'):
        print(item.name)
        for inner in item.glob('**/*'):
            if inner.is_file():
                get_suf = inner.suffix.lower()
                print(inner.name)
                if get_suf not in know_suffix:
                    for cat, exts in CATEGORIES.items():
                        if get_suf in exts:
                            know_suffix.append(get_suf)
                    if get_suf not in unknown_suffix and get_suf not in know_suffix:
                        unknown_suffix.append(get_suf)

    print(f'Known suffix: {know_suffix}')
    print(f'Unknown suffix: {unknown_suffix}')


def main():
    try:
        path = Path(input('Enter path of folder:'))
    except IndexError:
        return 'No path to folder'

    if not path.exists():
        return f'Folder with path {path} doesn"t exist'
    sort_folder(path)
    removeEmptyFolders(path, removeRoot=True)
    unpack_archive(path)
    get_results(path)
    return 'All ok'


if __name__ == "__main__":
    removeRoot = True
    print(main())



import logging
import re

from fuzzywuzzy import fuzz, process
import readline

import assistant_ostap.assistant_ostap.classes as classes
from assistant_ostap.assistant_ostap.handlers import commands
from assistant_ostap.assistant_ostap.notes import NoteBook


# Даний метод відповідає за автозаповнення команд. Якщо у консолі
# ввести частину команди та натиснути tab то команда доповниться.
# у разі, якщо більше однієї команди відповідають критеріям,
# повернеться список усіх доступних команд
def completer(text, state):
    if not text.isalpha():
        return None
    options = [cmd for cmd in commands.keys() if cmd.startswith(text.lower())]
    if not options:
        return None
    if state < len(options):
        return options[state]
    return None


def parse_command(user_input: str):
    # Даний регулярний вираз шукає команди, що складаються більше, ніж з одного слова.
    # Тобто це good bye, show all, del phone та del user.
    # Якщо користувач ввів одну із цих команд, командою вважатимуться перші два слова,
    # а аргументами - все, починаючи з третього. Я
    # кщо ж команда складаєтсья з одного слова(блок else),
    # то аргументами є все, починаючи з другого елементу
    match = re.search(
        r"^show\s|^good\s|^del\s|^sort\s|^change\s|^add\s", user_input.lower())
    try:
        if match:
            user_command = " ".join(user_input.split()[:2]).lower()
            command_arguments = user_input.split()[2:]
        else:
            user_command = user_input.split()[0].lower()
            command_arguments = user_input.split()[1:]
    except IndexError:
        return "Please enter a command name."

    # Тут обробляються випадки, коли користувач ввів нвідому команду.
    # Якщо ввід схожий на існуючу команду, наприклад, chanle, то
    # повернеться повідомлення: "Можливо Ви мали на увазі change"
    # Якщо ж жодна із команд не буде достатньо подібною до
    # вводу користувача(за це відповідає змінна match_ratio
    # та коефіцієнт 60, виведений експерементальним шляхом),
    # повернеться повідомлення про те що команду не знайдено.
    if user_command not in commands.keys():
        logging.basicConfig(level=logging.ERROR)
        best_match, match_ratio = process.extractOne(user_command,
                                                     commands.keys(),
                                                     scorer=fuzz.ratio)
        if match_ratio >= 60:
            return f"Command not found.\nPerhaps you meant '{best_match}'."
        else:
            return "Command not found.\nTo view all available commands, enter 'help'."
    else:
        return commands[user_command](*command_arguments)


def main():
    # Ці дві лінійки безпосередньо пов'язані з функцією completer.
    # Вони відповідають за те, при натисканні на яку кнопку відбуватиметься автодоповнення.
    readline.set_completer(completer)
    readline.parse_and_bind("tab: complete")
    print("How can I help you?")
    while True:
        user_input = input("Enter command: ")
        result = parse_command(user_input)

        if result:
            # Якщо повернули ітератор(тобто команда show all), проходимося по ньому в циклі,
            #  поступово показуючи записи
            if isinstance(result, classes.AddressBook) or isinstance(result, NoteBook):
                for page in result:
                    commands["clear"]()
                    print("\n".join([str(i) for i in page]))
                    user_input = input(
                        "Press 'q' to quit. Press any key to see the next page: ")
                    if user_input.lower() == "q":
                        break
            else:
                print(result)


if __name__ == "__main__":
    main()



import os
import platform
import sys
from rich.console import Console
from rich.table import Table
import assistant_ostap.assistant_ostap.classes as classes
from assistant_ostap.assistant_ostap.notes import NoteBook
from assistant_ostap.assistant_ostap.clean import main
import re


commands = {}

# Декоратор set_commands створений для наповнення словника commands
# Ключами є команда, котра передається у якості аргумента name та, за потреби,
# additional. Значеннями є функції, що виконуються при введенні команди


def set_commands(name, *additional):
    def inner(func):
        commands[name] = func
        for command in additional:
            commands[command] = func
    return inner


def input_error(func):
    def inner(*args):
        try:
            return func(*args)
        except (IndexError, ValueError):
            return "Enter all require arguments please.\nTo see more info type 'help'."
        except classes.WrongPhone:
            return "You tried to enter an invalid phone number. Please check the value and try again"
        except classes.WrongDate:
            return "Invalid date. Please enter birthday in format 'DD.MM.YYYY'."
        except classes.WrongEmail:
            return "Invalid email address. Please enter a correct email address."
        except classes.NoName:
            return "Name is empty. Please try again"
        except KeyError:
            return "Id not found. Please check the value and try again"
            
    # Рядок нижче потрібний для того, щоб пов'язати функції та їх рядки документації.
    # Це потрібно для функції help.
    inner.__doc__ = func.__doc__
    return inner

@set_commands("help")
@input_error
def help_command(*args):
    """Show all commands available."""
    # Даний метод виводить користувачу перелік усіх команд з описом.
    # Створюється рядок all_commands, що наповнюється у циклі.
    # command це ключі в словнику commands.
    # func - значення. func.__doc__ це рядок документації, що додатково
    # прив'язувався до функції у декораторі input_error
    table = Table(title="Commands",style="magenta",show_lines=True)
    table.add_column('Comand')
    table.add_column('Description')
    for command, func in commands.items():
        table.add_row(command, func.__doc__)
    console = Console()
    console.print(table)


@set_commands("add record")
@input_error
def add(*args):
    """Take as input username, phone number, birthday, address, email and add them to the base.
    If username already exist add phone number to this user."""
    name = input('Enter name:')
    if classes.Name.is_valid_name(name):
        name = classes.Name(name)
    else:
        raise classes.NoName
    # Два блоки if, розміщених нижче відповідають за правильність введення телефону
    # та дня народження. Якщо значення невалідне, викликається помилка, що потім обробляється
    # деораторот input_error
    phone_number = input('Enter phone number:') 
    if classes.Phone.is_valid_phone(phone_number):
        phone_number = classes.Phone(phone_number)
    else:
        raise classes.WrongPhone
    # У змінній data зберігається екземпляр класу AddressBook із записаними раніше контактами
    # Змінна name_exists показує, чи існує контакт з таким ім'ям у data
    data = classes.AddressBook.open_file("data.json")
    name_exists = bool(data.get(name.value))

    # Тут відбувається перевірка, чи ім'я вже є у списку контактів
    # Якщо контакт відсутній, створюється новий Record.
    # Якщо присутній - до існуючого екземпляру Record додається номер телефону
    if name_exists and phone_number:
        # Методи класу Record повертають повідомлення для користувача
        # Дані повідомлення записуються у змінну та повертаються з функції
        # для показу користувачу
        msg = data[name.value].add_phone(phone_number)
        return msg
    birthday = input('Enter birthday:')
    if classes.Birthday.is_valid_date(birthday):
        birthday = classes.Birthday(birthday)
    else:
        raise classes.WrongDate
    street = input('Enter street:')
    city = input('Enter city:')
    country = input('Enter country:')
    postcode = input('Enter postcode:')
    address = classes.Address(street, city, country, postcode)

    email_value = input('Enter email:')
    if not classes.Email.is_valid_email(email_value):
        raise classes.WrongEmail

    email = classes.Email(email_value)

    record = classes.Record(name, [phone_number], birthday, address, email)
    data.add_record(record)
    data.write_to_file("data.json")
    return f"User {name} added successfully."

    
@set_commands("add phone")
@input_error
def add_phone(*args):
    """Takes as input username, phone number and adds to the contact."""

    name = classes.Name(input('Enter name:'))
    phone = input('Enter phone:')
    if classes.Phone.is_valid_phone(phone):
        new_phone = classes.Phone(phone)
    else:
        raise classes.WrongPhone

    data = classes.AddressBook.open_file("data.json")
    name_exists = bool(data.get(name.value))
    if not name_exists:
        msg = f"Name {name} doesn't exist. "\
            "If you want to add it, please use add command."
    else:
        msg = data[name.value].add_phone(new_phone)

    data.write_to_file("data.json")
    return msg


@set_commands("add note")
@input_error
def create_note(*args):
    """Take as input the text of the note in quotes and adds it to the notebook"""
    text = " ".join(args)
    nb = NoteBook.read_from_file()
    nb.add_note(text)

    nb.save_to_file()
    return "Note added successfully."


@set_commands("change phone")
@input_error
def change(*args):
    """Take as input username, old and new phone number 
    and changes the corresponding data."""

    name = classes.Name(input('Enter name:'))
    old_phone = classes.Phone(input('Enter old phone:'))
    new_phone = input('Enter new phone:')
    if classes.Phone.is_valid_phone(new_phone):
        new_phone = classes.Phone(new_phone)
    else:
        raise classes.WrongPhone

    data = classes.AddressBook.open_file("data.json")
    name_exists = bool(data.get(name.value))

    if not name_exists:
        msg = f"Name {name} doesn`t exist. "\
            "If you want to add it, please type 'add user <name> <phone number>'."
    else:
        msg = data[name.value].change_phone(old_phone, new_phone)

    data.write_to_file("data.json")
    return msg


@set_commands("change birthday")
@input_error
def change_birthday(*args):
    """Takes as input username, birthday date
    and changes the corresponding data."""

    name = classes.Name(input('Enter name:'))
    birthday = input('Enter b-day:')
    if classes.Birthday.is_valid_date(birthday):
        new_birthday = classes.Birthday(birthday)
    else:
        raise classes.WrongDate(
                "Invalid date. Please enter birthday in format 'DD.MM.YYYY'.")

    data = classes.AddressBook.open_file("data.json")
    name_exists = bool(data.get(name.value))
    if not name_exists:
        msg = f"Name {name} doesn`t exist. "\
            "If you want to add it, please type 'add user <name> <phone number>'."
    else:
        msg = data[name.value].change_birthday(new_birthday)

    data.write_to_file("data.json")
    return msg


@set_commands("change address")
@input_error
def change_address(*args):
    """Takes as input username, new address and changes the corresponding data."""

    if len(args) < 5:
        return "Not valid command format."\
            "Please enter correct address data"
    name = classes.Name(input('Enter name:'))
    street = input('Enter street:')
    city = input('Enter city:')
    country = input('Enter country:')
    postcode = input('Enter postcode:')
    data = classes.AddressBook.open_file("data.json")
    name_exists = bool(data.get(name.value))
    if not name_exists:
        msg = f"Name {name} doesn't exist. "\
            "If you want to add it, please use add command."
    else:
        address = classes.Address(street, city,
                              country, postcode)
        msg = data[name.value].change_address(address)

    data.write_to_file("data.json")
    return msg


@set_commands("change email")
@input_error
def change_email(*args):
    """Takes as input username, new email and changes the corresponding data."""

    name = classes.Name(input('Enter name:'))
    data = classes.AddressBook.open_file("data.json")
    name_exists = bool(data.get(name.value))
    if not name_exists:
        msg = f"Name {name} doesn't exist. "\
            "If you want to add it, please use add command."
    else:
        email_value = input('Enter email:')
        if not classes.Email.is_valid_email(email_value):
            raise classes.WrongEmail
        new_email = classes.Email(email_value)
        msg = data[name.value].change_email(new_email)

    data.write_to_file("data.json")
    return msg


@set_commands("change note")
@input_error
def edit_note(*args):
    """Take as input note id and change selected note"""
    note_id = input('Enter note ID')
    nb = NoteBook.read_from_file()
    nb.edit_note(note_id)

    nb.save_to_file()
    return "Note edited successfully."


@set_commands("del user")
@input_error
def delete_user(*args):
    """Take as input username and delete that user"""
    name = classes.Name(input('Enter name:'))

    data = classes.AddressBook.open_file("data.json")
    name_exists = bool(data.get(name.value))

    if not name_exists:
        return f"Name {name} doesn`t exist."
    else:
        data.delete_record(name)

    data.write_to_file("data.json")
    return f"User {name} deleted successfully."


@set_commands("del phone")
@input_error
def delete_phone(*args):
    """Takes as input username and phone number and deletes that phone"""
    name = classes.Name(input('Enter name:'))
    phone = classes.Phone(input('Enter phone:'))

    data = classes.AddressBook.open_file("data.json")
    name_exists = bool(data.get(name.value))

    if not name_exists:
        msg = f"Name {name} doesn`t exist."
    else:
        msg = data[name.value].delete_phone(phone)

    data.write_to_file("data.json")
    return msg


@set_commands("del note")
@input_error
def del_note(*args):
    """Take the input username and show the address"""
    name = classes.Name(input('Enter name:'))
    data = classes.AddressBook.open_file("data.csv")
    name_exists = bool(data.get(name.value))

    if not name_exists:
        return f"Name {name} doesn't exist."

    else:
        email_str = str(data[name.value].email)

        if email_str:
            return f"Email for {name}: {email_str}."

        else:
            return f"There isn't email for user {name}."


@set_commands("show all")
@input_error
def show_all(*args):
    """Show all users or notes"""
    field = input('Enter type of fields (users or notes):').lower()
    # Код функції show_all має саме такий вигляд тому, що AddressBook
    # це ітератор
    if field not in ("users", "notes"):
        return f"Unknown field {field}. Please type 'users' or 'notes'"
    if field == "users":
        return classes.AddressBook.open_file("data.json")
    return NoteBook.read_from_file()


@set_commands("show phone")
@input_error
def phone(*args):
    """Take as input username and show user`s phone number."""
    name = classes.Name(input('Enter name:'))

    data = classes.AddressBook.open_file("data.json")
    name_exists = bool(data.get(name.value))

    if not name_exists:
        return f"Name {name} doesn`t exist. "\
            "If you want to add it, please type 'add <name> <phone number>'."
    else:
        # У цьому рядку список телефонів перетворюється у рядок,
        # де номері перелічені через кому
        phone_numbers = ", ".join(str(phone)
                                  for phone in data[name.value].phones)
        if phone_numbers:
            return f"Phone numbers for {name}: {phone_numbers}."
        else:
            return f"There are no phone numbers for user {name}"


@set_commands("show address")
@input_error
def address(*args):
    """Take the input username and show the address"""
    name = classes.Name(input('Enter name:'))

    data = classes.AddressBook.open_file("data.json")
    name_exists = bool(data.get(name.value))

    if not name_exists:
        return f"Name {name} doesn't exist"

    else:
        address_str = str(data[name.value].address)
        if address_str:
            return f"Address for {name}: {address_str}."

        else:
            return f"There is no address for user {name}."


@set_commands("show email")
@input_error
def email(*args):
    """Take the input username and show the email"""
    name = classes.Name(input('Enter name:'))
    data = classes.AddressBook.open_file("data.json")
    name_exists = bool(data.get(name.value))

    if not name_exists:
        return f"Name {name} doesn't exist."

    else:
        email_str = str(data[name.value].email)

        if email_str:
            return f"Email for {name}: {email_str}."

        else:
            return f"There isn't email for user {name}."


@set_commands("show nearbday")
@input_error
def show_birthdays_handler(*args):
    """Take as input number of days and show the list of birthdays.
    The MAX number of days is 365"""
    try:
        value = int(input('Enter count of days:'))
    except IndexError:
        return "Please enter the valid command: showbd number_of_days"
    if type(value) == int and value > 0:
        data = classes.AddressBook.open_file("data.json")
        if value > 365:
            value = 365
        return data.show_birthday(value)
    else:
        return "Please input a valid number of days"


@set_commands("search")
@input_error
def search_handler(*args):
    """Take as input searched field(name, phone, tag or text)
    and the text to be found. Returns all found users"""
    # у даній функції користувачу потрібно обрати, у яких полях
    # відбуватиметься пошук(наразі це name або phone) та ввести значення для пошуку.
    #  Функція повертає рядок з переліком усіх контаків
    field = input('Enter type of field to search by (name/phone/email/tag/text):')
    text = input('Enter value of field:')
    if field.lower() not in ("name", "phone", "email","tag","text"):
        return f"Unknown field '{field}'.\nTo see more info enter 'help'"

    nb = NoteBook.read_from_file()
    if field == "text":
        return nb.find_notes_by_text(text)
    elif field == "tag":
        return nb.find_notes_by_keyword(text)

    ab = classes.AddressBook.open_file("data.json")
    result = ab.search(field, text)
    if not result:
        return "There are no users matching"
    return "\n".join([str(rec) for rec in result])


@set_commands("sort notes")
@input_error
def sort_notes(*args):
    """Takes a keyword as input and sorts notes by it"""
    keyword = input('Enter tag what will be used for sorting notes:')
    nb = NoteBook.read_from_file()
    return nb.sort_notes(keyword)


@set_commands("sort files")
@input_error
def sort_files(*args):
    """Sort files by categories in input directory"""
    return main()


@set_commands("clear")
@input_error
def clear(*args):
    """Clear the console."""
    # Дана функція відповідальна за очищення консолі.
    # Darwin це macOS
    system = platform.system()
    if system == "Windows":
        os.system("cls")
    elif system in ("Linux", "Darwin"):
        os.system("clear")
    else:
        return "Sorry, this command is not available on your operating system."


@set_commands("exit", "close", "good bye", 'quit')
@input_error
def exit(*args):
    """Interrupt program."""
    sys.exit(0)

# Для того, щоб дадати нові команди до бота достатньо просто
# тут написати іх код з відповідними декораторами та docstring.
# Наприклад, уявімо, що потрібно додати команду, що повертатиме ввід користувача.
# Якщо розчоментувати кож нижче можна побачити, що команда echo працює успішно

# @set_commands("echo")
# @input_error
# def echo(*args):
#     """Return user`s input"""
#     return " ".join(args)



import setuptools

# Открытие README.md и присвоение его long_description.
with open("README.md", "r") as fh:
	long_description = fh.read()

# Определение requests как requirements для того, чтобы этот пакет работал. Зависимости проекта.
requirements = ['fuzzywuzzy==0.18.0',
                'Levenshtein==0.21.1',
                'pyreadline3==3.4.1',
                'python-Levenshtein==0.21.1',
                'rapidfuzz==3.1.1',
                'rich==13.5.2']

# Функция, которая принимает несколько аргументов. Она присваивает эти значения пакету.
setuptools.setup(
	# Имя дистрибутива пакета. Оно должно быть уникальным, поэтому добавление вашего имени пользователя в конце является обычным делом.
	name="assistant_ostap",
	# Номер версии вашего пакета. Обычно используется семантическое управление версиями.
	version="0.0.21",
	# Имя автора.
	author="GoIt Team 3",
	# Его почта.
	author_email="",
	# Краткое описание, которое будет показано на странице PyPi.
	description="Your personal assistant Ostap",
	# Длинное описание, которое будет отображаться на странице PyPi. Использует README.md репозитория для заполнения.
	long_description=long_description,
	# Определяет тип контента, используемый в long_description.
	long_description_content_type="text/markdown",
	# URL-адрес, представляющий домашнюю страницу проекта. Большинство проектов ссылаются на репозиторий.
	url="https://github.com/NeverInMind/Project_Team3/",
	# Находит все пакеты внутри проекта и объединяет их в дистрибутив.
	packages=setuptools.find_packages(),
	# requirements или dependencies, которые будут установлены вместе с пакетом, когда пользователь установит его через pip.
	install_requires=requirements,
	# Предоставляет pip некоторые метаданные о пакете. Также отображается на странице PyPi.
	classifiers=[
		"Programming Language :: Python :: 3.8",
		"License :: OSI Approved :: MIT License",
		"Operating System :: OS Independent",
	],
	# Требуемая версия Python.
	python_requires='>=3.6',
	include_package_data=True,
    entry_points={'console_scripts': ['Ostap = assistant_ostap.main:main']}
)


from collections import UserDict
from dataclasses import dataclass, asdict
import json
import random
import re

import readline


@dataclass
class Note:
    text: str
    id: str
    tags: list = list

    def __post_init__(self):
        self.tags = self.keywords

    @property
    def keywords(self) -> list:
        # Ключовим вважається слово, перед яким у тексті є знак "#"
        pattern = r"#(\w+)"
        tags = re.findall(pattern, self.text.lower())
        return tags

    def __str__(self):
        return f"{self.id}: {self.text}"

    def __repr__(self):
        return str(self)


class NoteBook(UserDict):
    def __add__(self, other):
        if isinstance(other, NoteBook):
            new_notebook = NoteBook()
            new_notebook.update(self.data)
            new_notebook.update(other.data)
            return new_notebook
        else:
            raise TypeError("Can only add two NoteBook instances together")

    def add_note(self, text: str):
        note_id = str(random.randint(1000, 9999))
        note = Note(text, note_id)
        self.data[note_id] = note

    def edit_note(self, note_id):
        def set_initial_input(text):
            def hook():
                readline.insert_text(text)
                readline.redisplay()
            readline.set_pre_input_hook(hook)

        set_initial_input(self.data[note_id].text)
        user_input = input()
        new_note = Note(user_input, id=note_id)
        self.data[note_id] = new_note
        set_initial_input("")

    def del_note(self, note_id):
        del self[note_id]

    def find_notes_by_keyword(self, keyword):
        result = []
        for note in self.data.values():
            if keyword in note.tags:
                result.append(str(note))
        if not result:
            return "There are no notes matching"
        return "\n".join(result)

    def find_notes_by_text(self, text):
        result = []
        for note in self.data.values():
            if text in note.text:
                result.append(str(note))
        if not result:
            return "There are no notes matching"
        return "\n".join(result)

    def sort_notes(self, keyword):
        notes_with_keyword = NoteBook()
        notes_without_keyword = NoteBook()
        nb = NoteBook.read_from_file()

        for note in nb.data.values():
            if keyword.lower() in note.tags:
                notes_with_keyword[note.id] = note
            else:
                notes_without_keyword[note.id] = note

        if not notes_with_keyword:
            return f"Keyword {keyword} not found"

        return notes_with_keyword + notes_without_keyword

    def save_to_file(self):
        result = {}
        for note_id, note in self.data.items():
            result[str(note_id)] = asdict(note)

        with open("notebook.json", "w") as file:
            json.dump(result, file, indent=4, ensure_ascii=False)

    @classmethod
    def read_from_file(cls):
        try:
            with open("notebook.json") as file:
                data_json = json.load(file)
                data = cls()
                for note_json in data_json.values():
                    note = Note(**note_json)
                    data[note.id] = note

        except FileNotFoundError:
            data = cls()
        return data

    def __iter__(self):
        self.current_page = 1
        self.page_size = 10
        self.start_index = (self.current_page - 1) * self.page_size
        self.end_index = self.start_index + self.page_size
        return self

    def __next__(self):
        if self.start_index >= len(self.data):
            raise StopIteration

        page_records = list(self.data.values())[
            self.start_index:self.end_index]
        self.start_index = self.end_index
        self.end_index = self.start_index + self.page_size
        self.current_page += 1

        return page_records
